/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import fr.x9c.cadmium.kernel.AbstractNativeRunner;
import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * Entry point of the Barista Java API.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.0
 */
public final class API {

    /** The actual Barista program (that registered callbacks to be used). */
    private static fr.x9c.barista.cafesterolMain main;

    /**
     * No instance of this class.
     */
    private API() {
    } // end empty constructor

    /**
     * Initializes the Barista API.
     */
    public static synchronized void init() {
        if (main == null) {
            main = fr.x9c.barista.cafesterolMain.mainWithReturn(new String[] { "-api" });
        } // end if
    } // end method 'init()'

    /**
     * Runs a callback, wrapping any exception as a Barista one.
     * @param name callback name - should not be <tt>null</tt>
     * @param args callback parameters - should not be <tt>null</tt>
     * @return the result of the callback execution
     * @throws BaristaException if any error occurs during callback
     */
    static Value callback(final String name, final Value... args)
        throws BaristaException {
        assert name != null : "null name";
        assert args != null : "null args";
        init();
        final AbstractNativeRunner runner = main.copy();
        try {
            return runner.callback(name, args);
        } catch (final Fail.Exception fe) {
            final Value v = fe.asValue(runner);
            throw new BaristaException(v.asBlock().get(1).asBlock().asString());
        } catch (final Fatal.Exception fe) {
            throw new BaristaException(fe.getMessage());
        } catch (final FalseExit fe) {
            return null;
        } catch (final CadmiumException ce) {
            throw new BaristaException(ce.toString());
        } // end try/catch
    } // end method 'callback(String, Value...)'

    /**
     * Creates an UTF8 instance from a string.
     * @param s string to create from - should not be <tt>null</tt>
     * @return the corresponding UTF8 instance
     */
    static Value createUTF8(final String s) {
        assert s != null : "null s";
        try {
            final Block b = Block.createString(s.getBytes("UTF-8"));
            return Value.createFromBlock(b);
        } catch (java.io.UnsupportedEncodingException uee) {
            throw new BaristaException("UTF8 unsupported by Java platform");
        } // and try/catch
    } // end method 'createUTF8(String)'

    /**
     * Converts an UTF8 value into a string.
     * @param v value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     */
    static String stringOfUTF8(final Value v) {
        assert v != null : "null v";
        try {
            return new String(v.asBlock().getBytes(), "UTF-8");
        } catch (java.io.UnsupportedEncodingException uee) {
            throw new BaristaException("UTF8 unsupported by Java platform");
        } // and try/catch
    } // end method 'stringOfUTF8(Value)'

    /**
     * Creates a class name from a string.
     * @param s string to create name from - should not be <tt>null</tt>
     * @return the corresponding class name
     * @throws BaristaException if passed string is not a valid class name
     */
    static Value createClassName(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Name.make_class",
                        createUTF8(s));
    } // end method 'createClassName(String)'

    /**
     * Creates a field name from a string.
     * @param s string to create name from - should not be <tt>null</tt>
     * @return the corresponding field name
     * @throws BaristaException if passed string is not a valid field name
     */
    static Value createFieldName(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Name.make_field",
                        createUTF8(s));
    } // end method 'createFieldName(String)'

    /**
     * Creates a method name from a string.
     * @param s string to create name from - should not be <tt>null</tt>
     * @return the corresponding method name
     * @throws BaristaException if passed string is not a valid method name
     */
    static Value createMethodName(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Name.make_method",
                        createUTF8(s));
    } // end method 'createMethodName(String)'

    /**
     * Converts a class name value into a string.
     * @param v classname value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfClassName(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Name.utf8_of_class", v));
    } // end method 'stringOfClassName(Value)'

    /**
     * Converts a field name value into a string.
     * @param v field name value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfFieldName(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Name.utf8_of_field", v));
    } // end method 'stringOfFieldName(Value)'

    /**
     * Converts a method name value into a string.
     * @param v method name value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfMethodName(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Name.utf8_of_method", v));
    } // end method 'stringOfMethodName(Value)'

    /**
     * Creates a field signature from a string.
     * @param s string to create signature from - should not be <tt>null</tt>
     * @return the corresponding signature
     * @throws BaristaException if passed string is not a valid field signature
     */
    static Value createFieldSignature(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Signature.make_field",
                        createUTF8(s));
    } // end method 'createFieldSignature(String)'

    /**
     * Converts a field signature value into a string.
     * @param v field signature value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfFieldSignature(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Signature.utf8_of_field_signature", v));
    } // end method 'stringOfFieldSignature(Value)'

    /**
     * Creates a method signature from a string.
     * @param s string to create signature from - should not be <tt>null</tt>
     * @return the corresponding signature
     * @throws BaristaException if passed string is not a valid method signature
     */
    static Value createMethodSignature(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Signature.make_method",
                        createUTF8(s));
    } // end method 'createMethodSignature(String)'

    /**
     * Converts a method signature value into a string.
     * @param v method signature value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfMethodSignature(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Signature.utf8_of_method_signature", v));
    } // end method 'stringOfMethodSignature(Value)'

    /**
     * Creates a class signature from a string.
     * @param s string to create signature from - should not be <tt>null</tt>
     * @return the corresponding signature
     * @throws BaristaException if passed string is not a valid class signature
     */
    static Value createClassSignature(final String s) throws BaristaException {
        assert s != null : "null s";
        return callback("fr.x9c.barista.api.Signature.make_class",
                        createUTF8(s));
    } // end method 'createClassSignature(String)'

    /**
     * Converts a class signature value into a string.
     * @param v class signature value to convert - should not be <tt>null</tt>
     * @return the corresponding string
     * @throws BaristaEception if an error occurs during conversion
     */
    static String stringOfClassSignature(final Value v) throws BaristaException {
        assert v != null : "null v";
        return stringOfUTF8(callback("fr.x9c.barista.api.Signature.utf8_of_class_signature", v));
    } // end method 'stringOfClassSignature(Value)'

    /**
     * Decodes a list of value from a value.
     * @param v value to decode - should not be <tt>null</tt>
     * @return the corresponding list
     */
    static List<Value> decodeList(final Value v) {
        assert v != null : "null v";
        final List<Value> res = new LinkedList<Value>();
        Value list = v;
        while (list != Value.EMPTY_LIST) {
            final Block b = list.asBlock();
            res.add(b.get(0));
            list = b.get(1);
        } // end while
        return res;
    } // end method 'decodeList(Value)'

    /**
     * Encodes a list into a value.
     * @param l list to encode - should not be <tt>null</tt>
     * @return the passed list as a value
     */
    static <T extends ToValue> Value encodeList(final List<T> l) {
        assert l != null : "null l";
        Value res = Value.EMPTY_LIST;
        final ListIterator<T> it = l.listIterator(l.size());
        while (it.hasPrevious()) {
            final Block cons = Block.createBlock(Block.TAG_CONS,
                                                 it.previous().toValue(),
                                                 res);
            res = Value.createFromBlock(cons);
        } // end while
        return res;
    } // end method 'encodeList(List<T>)'

    /**
     * Encodes a list into a value.
     * @param l list to encode - should not be <tt>null</tt>
     * @return the passed list as a value
     */
    static Value encodeValueList(final List<Value> l) {
        assert l != null : "null l";
        Value res = Value.EMPTY_LIST;
        final ListIterator<Value> it = l.listIterator(l.size());
        while (it.hasPrevious()) {
            final Block cons = Block.createBlock(Block.TAG_CONS,
                                                 it.previous(),
                                                 res);
            res = Value.createFromBlock(cons);
        } // end while
        return res;
    } // end method 'encodeValueList(List<Value>)'

    /**
     * Encodes a list into a value.
     * @param ll list to encode - should not be <tt>null</tt>
     * @return the passed list as a value
     */
    static <T extends ToValue> Value encodeListList(final List< List<T> > ll) {
        assert ll != null : "null ll";
        final List<Value> l = new LinkedList<Value>();
        for (List<T> i : ll) {
            l.add(encodeList(i));
        } // end for
        return encodeValueList(l);
    } // end method 'encodeListList(List< List<T> >)'

    /**
     * Tests whether two objects are either equal or both <tt>null</tt>.
     * @param obj1 first object to test
     * @param obj2 second object to test
     * @return <tt>true</tt> if passed objects are either equal or both <tt>null</tt>,
     *         <tt>false</tt> otherwise
     */
    static boolean nullOrEquals(final Object obj1, final Object obj2) {
        if (obj1 != null) {
            return obj2 != null ? obj1.equals(obj2) : false;
        } else {
            return obj2 == null;
        } // end if/else
    } // end method 'nullOrEquals(Object, Object)'

} // end class 'API'
