/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;

/**
 * This class implements an <a href="http://ant.apache.org">Ant</a> task
 * assembling a class. <br>
 * The task supports the following attributes:
 * <ul>
 *   <li> <tt>file</tt>: file to assemble<br>
 *        <i>file</i> - no default value, should be set
 *   <li> <tt>path</tt>: output path for assembled file<br>
 *        <i>directory</i> - no default value, should be set
 *   <li> <tt>target</tt>: version used for class encoding<br>
 *        <i>string</i> - default value is <tt>1.6</tt>
 * </ul>
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.0
 */
public final class AntAssembleTask extends Task {

    /** File to assemble. */
    private File file;

    /** Output path for assembled file. */
    private File path;

    /** Version used for class encoding. */
    private String target;

    /**
     * Creates a new task with default parameters.
     */
    public AntAssembleTask() {
        this.file = null;
        this.path = null;
        this.target = "1.6";
    } // end empty constructor

    /**
     * Sets the file to assemble.
     * @param f file to assemble
     */
    public void setFile(final File f) {
        this.file = f;
    } // end method 'setFile(File)'

    /**
     * Sets the output path for assembled file.
     * @param f output path for assembled file
     */
    public void setPath(final File f) {
        this.path = f;
    } // end method 'setPath(File)'

    /**
     * Sets the version used for class encoding.
     * @param s version used for class encoding
     */
    public void setTarget(final String s) {
        this.target = s;
    } // end method 'setTarger(String)'

    /**
     * Actually executes the task.
     */
    public void execute() throws BuildException {
        if (this.file == null) {
            throw new BuildException("'file' attribute is not set.",
                                     getLocation());
        } // end if
        if (this.path == null) {
            throw new BuildException("'path' attribute is not set.",
                                     getLocation());
        } // end if
        final Version version = Version.fromString(this.target);
        if (version == null) {
            throw new BuildException("'target' attribute is invalid.",
                                     getLocation());
        } // end if

        try {
            Assembler.assemble(version,
                               new FileInputStream(this.file),
                               this.path.getAbsolutePath());
        } catch (final IOException ioe) {
            throw new BuildException(ioe.getMessage(), ioe, getLocation());
        } catch (final BaristaException be) {
            throw new BuildException(be.getMessage(), be, getLocation());
        } // end try/catch
    } // end method 'execute()'

} // end class 'AntAssembleTask'
