/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.io.InputStream;
import java.io.OutputStream;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.Custom;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides methods for class assembling from textual source.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.0
 */
public final class Assembler {

    /**
     * No instance of this class.
     */
    private Assembler() {
    } // end empty constructor

    /**
     * Calls the callback to assemble a class.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param computeStacks whether stack elements should be computed
     * @param classPath class path (used for stack computations) - should not be <tt>null</tt>
     * @param in where to read source - should not be <tt>null</tt>
     * @param dest where to output bytecode - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    private static String callAssemble(final Version v,
                                       final boolean computeStacks,
                                       final String classPath,
                                       final InputStream in,
                                       final Value dest)
        throws BaristaException {
        assert v != null : "null v";
        assert classPath != null : "null classPath";
        assert in != null : "null in";
        assert dest != null : "null dest";
        final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                           Custom.CHANNEL_OPS);
        b.setCustom(new Channel(in));
        final Value i = Value.createFromBlock(b);
        final Value res =
            API.callback("fr.x9c.barista.api.Assembler.assemble",
                         v.toValue(),
                         computeStacks ? Value.TRUE : Value.FALSE,
                         Value.createFromBlock(Block.createString(classPath)),
                         i,
                         dest);
        return API.stringOfUTF8(res);
    } // end method 'callAssemble(Version, boolean, String, InputStream, Value)'

    /**
     * Assembles from a textual source.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param computeStacks whether stack elements should be computed
     * @param classPath class path (used for stack computations) - should not be <tt>null</tt>
     * @param in where to read source - should not be <tt>null</tt>
     * @param out where to output bytecode - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final Version v,
                                  final boolean computeStacks,
                                  final String classPath,
                                  final InputStream in,
                                  final OutputStream out)
        throws BaristaException {
        final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                           Custom.CHANNEL_OPS);
        b.setCustom(new Channel(out));
	final Block c = Block.createBlock(1, Value.createFromBlock(b));
        final Block d = Block.createBlock(0, Value.createFromBlock(c));
        return callAssemble(v, computeStacks, classPath, in, Value.createFromBlock(d));
    } // end method 'assemble(Version, boolean, String, InputStream, OutputStream)'

    /**
     * Assembles from a textual source.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param in where to read source - should not be <tt>null</tt>
     * @param out where to output bytecode - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final Version v,
                                  final InputStream in,
                                  final OutputStream out)
        throws BaristaException {
        final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                           Custom.CHANNEL_OPS);
        b.setCustom(new Channel(out));
	final Block c = Block.createBlock(1, Value.createFromBlock(b));
        final Block d = Block.createBlock(0, Value.createFromBlock(c));
        return callAssemble(v, false, ".", in, Value.createFromBlock(d));
    } // end method 'assemble(Version, InputStream, OutputStream)'

    /**
     * Assembles from a textual source.
     * @param in where to read source - should not be <tt>null</tt>
     * @param out where to output bytecode - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final InputStream in,
                                  final OutputStream out)
        throws BaristaException {
        return assemble(Version.Java_1_6, in, out);
    } // end method 'assemble(InputStream, OutputStream)'

    /**
     * Assembles from a textual source.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param computeStacks whether stack elements should be computed
     * @param classPath class path (used for stack computations) - should not be <tt>null</tt>
     * @param in where to read source - should not be <tt>null</tt>
     * @param path where to output bytecode (needed subdirectories will be created)
     *             - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final Version v,
                                  final boolean computeStacks,
                                  final String classPath,
                                  final InputStream in,
                                  final String path)
        throws BaristaException {
        final Block b = Block.createString(path);
        final Block d = Block.createBlock(1, Value.createFromBlock(b));
        return callAssemble(v, computeStacks, classPath, in, Value.createFromBlock(d));
    } // end method 'assemble(Version, boolean, String, InputStream, String)'

    /**
     * Assembles from a textual source.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param in where to read source - should not be <tt>null</tt>
     * @param path where to output bytecode (needed subdirectories will be created)
     *             - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final Version v,
                                  final InputStream in,
                                  final String path)
        throws BaristaException {
        final Block b = Block.createString(path);
        final Block d = Block.createBlock(1, Value.createFromBlock(b));
        return callAssemble(v, false, ".", in, Value.createFromBlock(d));
    } // end method 'assemble(Version, InputStream, String)'

    /**
     * Assembles from a textual source.
     * @param in where to read source - should not be <tt>null</tt>
     * @param path where to output bytecode (needed subdirectories will be created)
     *             - should not be <tt>null</tt>
     * @return the name of the assembled class
     * @throws BaristaException if an errors occurs while assembling
     */
    public static String assemble(final InputStream in,
                                  final String path)
        throws BaristaException {
        return assemble(Version.Java_1_6, in, path);
    } // end method 'assemble(InputStream, String)'
    
} // end class 'Assembler'
