/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a classpath used to locate class files. <br>
 * <b>Such a classpath does not include the classpath of the running Java
 * program. This means that an empty classpath will be unable to find any
 * class, even the <tt>java.lang.Object</tt> class.</b>
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class ClassPath implements ToValue {

    /** Classpath elements. */
    private final List<String> elements;

    /**
     * Constructs an empty class path.
     */
    public ClassPath() {
        this.elements = new LinkedList<String>();
    } // end empty constructor

    /**
     * Constructs a classpath with one element.
     * @param elem classpath element - should not be <tt>null</tt>
     */
    public ClassPath(final String elem) {
        assert elem != null : "null elem";
        this.elements = new LinkedList<String>();
        this.elements.add(elem);
    } // end constructor(String)

    /**
     * Constructs a classpath from a list of elements.
     * @param elems classpath elements - should not be <tt>null</tt>
     */
    public ClassPath(final String... elems) {
        assert elems != null : "null elems";
        this.elements = new LinkedList<String>();
        for (String s : elems) {
            this.elements.add(s);
        } // end for
    } // end constructor(String...)

    /**
     * Constructs a classpath from a collection of elements.
     * @param elems classpath elements - should not be <tt>null</tt>
     */
    public ClassPath(final Collection<String> elems) {
        this.elements = new LinkedList<String>(elems);
    } // end constructor(Collection<String>)

    /**
     * Converts the classpath into a list of elements.
     * @return the classpath as a list of elements
     */
    public List<String> toList() {
        return new LinkedList<String>(this.elements);
    } // end method 'toList()'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder();
        boolean first = true;
        for (String s : this.elements) {
            if (first) {
                first = false;
            } else {
                sb.append(":");
            } // end if/else
            sb.append(s);
        } // end for
        return sb.toString();
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.elements.hashCode();
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof ClassPath) {
            final ClassPath that = (ClassPath) obj;
            return this.elements.equals(that.elements);
        } else {
            return false;
        } // end if/else
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        return Value.createFromBlock(Block.createString(toString()));
    } // end method 'toValue()'
    
} // end class 'ClassPath'
