(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module defines attributes in both low- and high-level forms.
   It also provides conversion functions between levels as well as i/o
   functions for low-level. *)


(** {6 Low-level form} *)

type info = {
    name_index : Utils.u2;
    length : Utils.u4;
    data : string;
  }
(** Represents an attribute as defined in the class file format specification. *)


(** {6 Exception} *)

type error =
  | Invalid_code_attribute
  | Invalid_code_length
  | Defined_twice of string
  | Invalid_field_attribute
  | Invalid_method_attribute
  | Invalid_class_attribute
  | Invalid_constant_value
  | Invalid_enclosing_method
  | Invalid_local_variable_table
  | Invalid_local_variable_type_table
  | Invalid_list_length
  | Invalid_attribute_name_value
  | Invalid_exception_name
  | Invalid_exception
  | Invalid_inner_class
  | Invalid_outer_class
  | Invalid_signature
  | Invalid_source_file
  | Invalid_stack_map_frame
  | Invalid_stack_map_verification_type

exception Exception of error
(** Exception to be raised when a function of this module fails.  *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 I/O functions} *)

val read_info : InputStream.t -> info
(** [read_info st] reads an attribute from [st].
   Raises [InputStream.Exception] if an i/o error occurs. *)

val write_info : OutputStream.t -> info -> unit
(** [write_info st a] writes an attribute [a] onto [st].
   Raises [OutputStream.Exception] if an i/o error occurs. *)


(** {6 High-level form} *)

type constant_value =
  | Long_value of int64 (** long constant *)
  | Float_value of float (** float constant *)
  | Double_value of float (** double constant *)
  | Boolean_value of bool (** boolean constant *)
  | Byte_value of int (** byte constant *)
  | Character_value of int (** char constant *)
  | Short_value of int (** short constant *)
  | Integer_value of int32 (** int constant *)
  | String_value of Utils.utf8 (** string constant *)
(** Constant values used by ConstantValue attribute. *)

type verification_type_info =
  | Top_variable_info (** {i top} type. *)
  | Integer_variable_info (** {i int} type. *)
  | Float_variable_info (** {i float} type. *)
  | Long_variable_info (** {i long} type. *)
  | Double_variable_info (** {i double} type. *)
  | Null_variable_info (** {i null} value type. *)
  | Uninitialized_this_variable_info (** uninitialized type (for {i this}). *)
  | Object_variable_info of [`Class_or_interface of Name.for_class | `Array_type of Descriptor.array_type] (** {i Object} type (parameter is the class name). *)
  | Uninitialized_variable_info of Utils.u2 (** uninitialized type (parameter is the code offset of the {i new} instruction that created the object). *)
(** Verification types. *)

type stack_map_frame =
  | Same_frame of Utils.u2 (** Same frame (parameter is code offset). *)
  | Same_locals_1_stack_item_frame of Utils.u2 * verification_type_info (** Same locals with one stack item (parameters are code offset and type of stack item). *)
  | Chop_1_frame of Utils.u2 (** Frame choped by 1 local element, operand stack is empty (parameter is code offset). *)
  | Chop_2_frame of Utils.u2 (** Frame choped by 2 local elements, operand stack is empty (parameter is code offset). *)
  | Chop_3_frame of Utils.u2 (** Frame choped by 3 local elements, operand stack is empty (parameter is code offset). *)
  | Append_1_frame of Utils.u2 * verification_type_info (** Frame appended by 1 local element, operand stack is empty (parameters are code offset and type of additional element). *)
  | Append_2_frame of Utils.u2 * verification_type_info * verification_type_info (** Frame appended by 2 local elements, operand stack is empty (parameters are code offset and types of additional elements). *)
  | Append_3_frame of Utils.u2 * verification_type_info * verification_type_info * verification_type_info (** Frame appended by 3 local elements, operand stack is empty (parameters are code offset and types of additional elements). *)
  | Full_frame of Utils.u2 * (verification_type_info list) * (verification_type_info list) (** Full frame definition (parameters are code offset and lists of locals and stack elements). *)
(** Stack map frame modifications. *)

type t =
  [ `ConstantValue of constant_value (** attribute for field initialization *)
  | `Code of code_value (** attribute for method code *)
  | `Exceptions of Name.for_class list (** attribute for thrown exceptions *)
  | `InnerClasses of ((Name.for_class option) * (Name.for_class option) * (Utils.utf8 option) * (AccessFlag.for_inner_class list)) list (** inner classes as a list of (inner class, outer class, inner name, access flags) *)
  | `EnclosingMethod of Name.for_class * ((Name.for_method * Descriptor.for_method) option) (** inner-most class and inner-most method, [None] if not immediatly enclosed by a method *)
  | `Synthetic (** auto-generated element *)
  | `Signature of signature (** element signature *)
  | `SourceFile of Utils.utf8 (** source file for class *)
  | `SourceDebugExtension of Utils.utf8 (** implementation specific *)
  | `LineNumberTable of (Utils.u2 * Utils.u2) list (** (code address, source line) mapping *)
  | `LocalVariableTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Descriptor.for_field * Utils.u2) list (** types for local variables *)
  | `LocalVariableTypeTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Signature.field_type_signature * Utils.u2) list (** signatures for local variables *)
  | `Deprecated (** deprecated elements *)
  | `RuntimeVisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeInvisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeVisibleParameterAnnotations of Annotation.t list list (** annotations *)
  | `RuntimeInvisibleParameterAnnotations of Annotation.t list list (** annotations *)
  | `AnnotationDefault of Annotation.element_value (** annotations default value *)
  | `StackMapTable of stack_map_frame list (** stack map table *)
  | `Unknown of Utils.utf8 * string (** unknown attribute *) ]
and signature =
  [ `Class of Signature.class_signature (** Signature for a class. *)
  | `Method of Signature.method_signature (** Signature for a method. *)
  | `Field of Signature.field_type_signature (** Signature for a field. *) ]
and code_attribute = [
  | `LineNumberTable of (Utils.u2 * Utils.u2) list (** line numbers *)
  | `LocalVariableTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Descriptor.for_field * Utils.u2) list (** types for local variables *)
  | `LocalVariableTypeTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Signature.field_type_signature * Utils.u2) list (** signatures for local variables *)
  | `StackMapTable of stack_map_frame list (** stack map table *)
  | `Unknown of Utils.utf8 * string (** unknown attribute *) ]
and code_value = {
    max_stack : Utils.u2;
    max_locals : Utils.u2;
    code : Instruction.t list;
    exception_table : (Utils.u2 * Utils.u2 * Utils.u2 * (Name.for_class option)) list;
    attributes : code_attribute list;
  }

type for_field =
  [ `ConstantValue of constant_value (** attribute for field initialization *)
  | `Synthetic (** auto-generated element *)
  | `Signature of [`Field of Signature.field_type_signature] (** element signature *)
  | `Deprecated (** deprecated elements *)
  | `RuntimeVisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeInvisibleAnnotations of Annotation.t list (** annotations *)
  | `Unknown of Utils.utf8 * string (** unknown attribute *) ]
(** Possible attributes for a field. *)

type for_method =
  [ `Code of code_value (** attribute for method code *)
  | `Exceptions of Name.for_class list (** attribute for thrown exceptions *)
  | `Synthetic (** auto-generated element *)
  | `Signature of [`Method of Signature.method_signature] (** element signature *)
  | `Deprecated (** deprecated elements *)
  | `RuntimeVisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeInvisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeVisibleParameterAnnotations of Annotation.t list list (** annotations *)
  | `RuntimeInvisibleParameterAnnotations of Annotation.t list list (** annotations *)
  | `AnnotationDefault of Annotation.element_value (** annotations default value *)
  | `Unknown of Utils.utf8 * string (** unknown attribute *) ]
(** Possible attributes for a method. *)

type for_class =
  [ `InnerClasses of ((Name.for_class option) * (Name.for_class option) * (Utils.utf8 option) * (AccessFlag.for_inner_class list)) list (** inner classes as a list of (inner class, outer class, inner name, access flags) *)
  | `EnclosingMethod of Name.for_class * ((Name.for_method * Descriptor.for_method) option) (** inner-most class and inner-most method, [None] if not immediatly enclosed by a method *)
  | `Synthetic (** auto-generated element *)
  | `Signature of [`Class of Signature.class_signature] (** element signature *)
  | `SourceFile of Utils.utf8 (** source file for class *)
  | `SourceDebugExtension of Utils.utf8 (** implementation specific *)
  | `Deprecated (** deprecated elements *)
  | `RuntimeVisibleAnnotations of Annotation.t list (** annotations *)
  | `RuntimeInvisibleAnnotations of Annotation.t list (** annotations *)
  | `Unknown of Utils.utf8 * string (** unknown attribute *) ]
(** Possible attributes for a class. *)

type element =
  | Class (** Kind of classes. *)
  | Method (** Kind of methods. *)
  | Field (** Kind of fields. *)
(** Kinds of elements. *)


(** {6 Conversion functions} *)

val check_code_attributes : t list -> [`LineNumberTable of (Utils.u2 * Utils.u2) list | `LocalVariableTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Descriptor.for_field * Utils.u2) list | `LocalVariableTypeTable of (Utils.u2 * Utils.u2 * Utils.utf8 * Signature.field_type_signature * Utils.u2) list | `StackMapTable of stack_map_frame list | `Unknown of Utils.utf8 * string] list
(** Checks that the passed list is a valid one for a code attribute.
   The list is returned if it is valid, raising [Exception] otherwise. *)

val check_field_attributes : t list -> for_field list
(** Checks that the passed list is a valid one for a field.
   The list is returned if it is valid, raising [Exception] otherwise. *)

val check_method_attributes : t list -> for_method list
(** Checks that the passed list is a valid one for a method.
   The list is returned if it is valid, raising [Exception] otherwise. *)

val check_class_attributes : t list -> for_class list
(** Checks that the passed list is a valid one for a class.
   The list is returned if it is valid, raising [Exception] otherwise. *)

val decode : element -> ConstantPool.t -> info -> t
(** Converts from a low-level into a high-level form according to passed pool.
   Raises [Exception] if an error occurs during conversion. *)

val encode : ConstantPool.extendable -> t -> info
(** Converts from a high-level into a low-level form, using passed pool.
   Raises [Exception] if an error occurs during conversion. *)

val version_bounds : t -> Version.bounds
(** Returns the version bounds for the passed attribute. *)
