(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Type *)

type t = {
    access_flags : AccessFlag.for_class list;
    name : Name.for_class;
    extends : Name.for_class option;
    implements : Name.for_class list;
    fields : Field.t list;
    methods : Method.t list;
    attributes : Attribute.for_class list;
  }


(* Exception *)

type error =
  | Invalid_class_name
  | Too_many of string
  | Version_error of Version.error

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_class_name -> "invalid class name"
  | Too_many s -> "too many " ^ s
  | Version_error e -> Version.string_of_error e

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Traversal *)

let map mapper cd =
  let flags', name', extends', implements', fields', methods', attributes' =
    mapper#class_definition
      cd.access_flags
      cd.name
      cd.extends
      cd.implements
      cd.fields
      cd.methods
      cd.attributes in
  { access_flags = flags';
    name = name';
    extends = extends';
    implements = implements';
    fields = fields';
    methods = methods';
    attributes = attributes' }

let iter iterator cd =
  iterator#class_definition
    cd.access_flags
    cd.name
    cd.extends
    cd.implements
    cd.fields
    cd.methods
    cd.attributes

let version_check_iterator v =
  let check_flags =
    List.iter (fun x -> Version.check (AccessFlag.version_bounds x) v) in
  let check_attribute x =
    Version.check (Attribute.version_bounds x) v in
  object
    inherit Traversal.default_class_definition_iterator
    method class_flags l =
      check_flags (l :> AccessFlag.t list)
    method field_flags l =
      check_flags (l :> AccessFlag.t list)
    method regular_method_flags l =
      check_flags (l :> AccessFlag.t list)
    method constructor_method_flags l =
      check_flags (l :> AccessFlag.t list)
    method initializer_method_strictfp b =
      if b then check_flags [`Strict]
    method class_attribute a =
      check_attribute (a :> Attribute.t)
    method field_attribute a =
      check_attribute (a :> Attribute.t)
    method regular_method_attribute a =
      check_attribute (a :> Attribute.t)
    method constructor_method_attribute a =
      check_attribute (a :> Attribute.t)
    method initializer_method_attribute a =
      check_attribute (a :> Attribute.t)
  end


(* Conversion functions *)

let decode cf =
  let pool = cf.ClassFile.constant_pool in
  let get_class_name idx = match ConstantPool.get_entry pool idx with
  | ConstantPool.Class idx' ->
      (match ConstantPool.get_entry pool idx' with
      | ConstantPool.UTF8 n ->
          Name.make_for_class_from_internal n
      | _ -> fail Invalid_class_name)
  | _ -> fail Invalid_class_name in
  let flags = AccessFlag.check_class_flags (AccessFlag.from_u2 false cf.ClassFile.access_flags) in
  let is_interface = List.mem `Interface flags in
  { access_flags = flags;
    name = get_class_name cf.ClassFile.this_class;
    extends =
    if cf.ClassFile.super_class = 0 then
      None
    else
      Some (get_class_name cf.ClassFile.super_class);
    implements = List.map
      get_class_name
      (Array.to_list cf.ClassFile.interfaces);
    fields = List.map
      (Field.decode is_interface pool)
      (Array.to_list cf.ClassFile.fields);
    methods = List.map
      (Method.decode is_interface pool)
      (Array.to_list cf.ClassFile.methods);
    attributes = Attribute.check_class_attributes
      (List.map
         (fun x -> Attribute.decode Attribute.Class pool x)
         (Array.to_list cf.ClassFile.attributes)); }

let encode ?(version=Version.Java_1_6) cd =
  let checked_length s arr =
    let res = Array.length arr in
    if res < 65536 then
      res
    else
      fail (Too_many s) in
  (try
    iter (version_check_iterator version) cd
  with Version.Exception e -> fail (Version_error e));
  let major, minor = Version.major_minor_of_version version in
  let pool = ConstantPool.make_extendable () in
  let this_index = ConstantPool.add_class pool cd.name in
  let super_index = match cd.extends with
  | Some n -> ConstantPool.add_class pool n
  | None -> 0 in
  let itfs = Array.of_list (List.map (fun s -> ConstantPool.add_class pool s) cd.implements) in
  let flds = Array.of_list (List.map (Field.encode pool) cd.fields) in
  let mths = Array.of_list (List.map (Method.encode pool) cd.methods) in
  let atts = Array.of_list (List.map (Attribute.encode pool) (cd.attributes :> Attribute.t list)) in
  let cpool = ConstantPool.to_pool pool in
  { ClassFile.magic = magic_number;
    ClassFile.minor_version = minor;
    ClassFile.major_version = major;
    ClassFile.constant_pool_count = checked_length "constant pool elements" cpool;
    ClassFile.constant_pool = cpool;
    ClassFile.access_flags = AccessFlag.list_to_u2 (cd.access_flags :> AccessFlag.t list);
    ClassFile.this_class = this_index;
    ClassFile.super_class = super_index;
    ClassFile.interfaces_count = checked_length "interfaces" itfs;
    ClassFile.interfaces = itfs;
    ClassFile.fields_count = checked_length "fields" flds;
    ClassFile.fields = flds;
    ClassFile.methods_count = checked_length "methods" mths;
    ClassFile.methods = mths;
    ClassFile.attributes_count = checked_length "attributes" atts;
    ClassFile.attributes = atts; }
