(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Constants *)

let dot_class = utf8_of_string ".class "

let dot_method = utf8_of_string ".method "

let dot_field = utf8_of_string ".field "

let dot_implements = utf8_of_string ".implements "

let dot_extends = utf8_of_string ".extends "

let space = utf8_of_string " "

let tab = utf8_of_string "  "

let colon = utf8_of_string ":"

let comma = utf8_of_string ","

let dot = utf8_of_string "."

let opening_parenthesis = utf8_of_string "("

let closing_parenthesis = utf8_of_string ")"

let tilde = utf8_of_string " ~ "

let zero = utf8_of_string "0"

let wide = utf8_of_string "wide "

let at_constant_value = utf8_of_string "@ConstantValue "

let at_unknown = utf8_of_string "@Unknown "

let at_exceptions = utf8_of_string "@Exceptions "

let at_inner_class = utf8_of_string "@InnerClass "

let at_enclosing_method = utf8_of_string "@EnclosingMethod "

let at_synthetic = utf8_of_string "@Synthetic"

let at_signature = utf8_of_string "@Signature "

let at_source_file = utf8_of_string "@SourceFile "

let at_source_debug_extension = utf8_of_string "@SourceDebugExtension "

let at_deprecated = utf8_of_string "@Deprecated"

let at_runtime_visible_annotations = utf8_of_string "@RuntimeVisibleAnnotations "

let at_runtime_invisible_annotations = utf8_of_string "@RuntimeInvisibleAnnotations "

let at_annotation_default = utf8_of_string "@AnnotationDefault"

let class_constructor_name = Name.make_for_method class_constructor

let class_initializer_name = Name.make_for_method class_initializer


(* Functions *)

let utf8_of_method_desc name desc =
  let params, return = desc in
  (Source.utf8_of_java_type return)
    ++ space
    ++ (Name.utf8_for_method name)
    ++ opening_parenthesis
    ++ (utf8_concat_sep comma (List.map Source.utf8_of_java_type (params :> Descriptor.java_type list)))
    ++ closing_parenthesis

let utf8_of_method_call name desc =
  let params, return = desc in
  (Name.utf8_for_method name)
    ++ opening_parenthesis
    ++ (utf8_concat_sep comma (List.map Source.utf8_of_java_type (params :> Descriptor.java_type list)))
    ++ closing_parenthesis
    ++ colon
    ++ (Source.utf8_of_java_type return)

let extract_code_attributes l =
  let rec extract lnt lvt lvtt st unk l =
    match l with
    | hd :: tl ->
        (match hd with
        | `LineNumberTable l -> extract (lnt @ l) lvt lvtt st unk tl
        | `LocalVariableTable l -> extract lnt (lvt @ l) lvtt st unk tl
        | `LocalVariableTypeTable l -> extract lnt lvt (lvtt @ l) st unk tl
        | `StackMapTable l -> extract lnt lvt lvtt l unk tl
        | `Unknown (n, s) -> extract lnt lvt lvtt st ((n, s) :: unk) tl)
    | [] ->
        ((List.sort (fun (x, _) (y, _) -> compare x y) lnt),
         (List.sort (fun (x, _, _, _, _) (y, _, _, _, _) -> compare x y) lvt),
         (List.sort (fun (x, _, _, _, _) (y, _, _, _, _) -> compare x y) lvtt),
         st,
         (List.rev unk)) in
  extract [] [] [] [] [] l

let rec annotation_value_list prefix = function
  | Annotation.Boolean_value b ->
      if b then
        [prefix ++ (utf8_of_string " boolean 1")]
      else
        [prefix ++ (utf8_of_string " boolean 0")]
  | Annotation.Byte_value b ->
      [prefix ++ (utf8_of_string (Printf.sprintf " byte %d" b))]
  | Annotation.Char_value c ->
      [prefix ++ (utf8_of_string (Printf.sprintf " char %d" (code_of_uchar c)))]
  | Annotation.Double_value d ->
      [prefix ++ (utf8_of_string (Printf.sprintf " double %f" d))]
  | Annotation.Float_value f ->
      [prefix ++ (utf8_of_string (Printf.sprintf " float %f" f))]
  | Annotation.Int_value i ->
      [prefix ++ (utf8_of_string (Printf.sprintf " int %ld" i))]
  | Annotation.Long_value l ->
      [prefix ++ (utf8_of_string (Printf.sprintf " long %Ld" l))]
  | Annotation.Short_value s ->
      [prefix ++ (utf8_of_string (Printf.sprintf " short %d" s))]
  | Annotation.String_value s ->
      [prefix ++ (utf8_of_string " string ") ++ (utf8_escape s)]
  | Annotation.Enum_value (cn, fn) ->
      [prefix
         ++ (utf8_of_string " enum ")
         ++ (Name.external_utf8_for_class cn)
         ++ space
         ++ (Name.utf8_for_field fn)]
  | Annotation.Class_value cn ->
      [prefix ++ (utf8_of_string " class ") ++ (Name.external_utf8_for_class cn)]
  | Annotation.Annotation_value a ->
      let prefix' = prefix ++ (utf8_of_string " annotation") in
      annotation_list prefix' a
  | Annotation.Array_value l ->
      let idx = ref 0 in
      List.flatten
        (List.map
           (fun x ->
             let prefix' = prefix
                 ++ space
                 ++ (utf8_of_string (string_of_int !idx)) in
             incr idx;
             annotation_value_list prefix' x)
           l)
and annotation_list prefix (name, pairs) =
  let prefix' = prefix
      ++ (Name.external_utf8_for_class name)
      ++ space in
  if pairs = [] then
    [prefix']
  else
    List.flatten
      (List.map
         (fun (id, vl) ->
           annotation_value_list (prefix' ++ id) vl)
         pairs)

let add_attribute buffer a =
  utf8_buf_add_string buffer tab;
  match a with
  | `ConstantValue cv ->
      utf8_buf_add_string buffer at_constant_value;
      (match cv with
      | Attribute.Long_value lv ->
          utf8_buf_add_endline buffer (utf8_of_string (Int64.to_string lv))
      | Attribute.Float_value fv ->
          utf8_buf_add_endline buffer (utf8_of_string (string_of_float fv))
      | Attribute.Double_value dv ->
          utf8_buf_add_endline buffer (utf8_of_string (string_of_float dv))
      | Attribute.Boolean_value bv ->
          utf8_buf_add_endline
            buffer
            (utf8_of_string (if bv then "1" else "0"))
      | Attribute.Byte_value bv ->
          utf8_buf_add_endline buffer (utf8_of_string (string_of_int bv))
      | Attribute.Character_value cv ->
          utf8_buf_add_endline buffer (utf8_of_string (string_of_int cv))
      | Attribute.Short_value sv ->
          utf8_buf_add_endline buffer (utf8_of_string (string_of_int sv))
      | Attribute.Integer_value iv ->
          utf8_buf_add_endline buffer (utf8_of_string (Int32.to_string iv))
      | Attribute.String_value sv ->
          utf8_buf_add_endline
            buffer
            (utf8_escape sv))
  | `Code cv ->
      let line_number_table, local_variable_table, local_variable_type_table, stack_map_table, unknowns =
        extract_code_attributes cv.Attribute.attributes in
      let prev = ref (-1) in
      let ofs = ref 0 in
      utf8_buf_add_endline
        buffer
        (utf8_of_string
           (Printf.sprintf ".max_stack %d" cv.Attribute.max_stack));
      utf8_buf_add_endline
        buffer
        (utf8_of_string
           (Printf.sprintf "  .max_locals %d" cv.Attribute.max_locals));
      List.iter
        (fun i ->
          (try
            let line = snd (List.find (fun (x, _) -> !prev < x && !ofs >= x) line_number_table) in
            utf8_buf_add_endline buffer (utf8_of_string (Printf.sprintf "  @LineNumberTable %d" line));
          with Not_found -> ());
          List.iter
            (fun (start, length, name, desc, index) ->
              let finish = start + length in
              if (!ofs >= start && !ofs < finish)
                  && not (!prev >= start && !prev < finish) then
                begin
                  utf8_buf_add_string buffer (utf8_of_string (Printf.sprintf "  @LocalVariableTable code%08d: code%08d: " start finish));
                  utf8_buf_add_string buffer name;
                  utf8_buf_add_string buffer space;
                  utf8_buf_add_string buffer (Source.utf8_of_java_type (desc :> Descriptor.java_type));
                  utf8_buf_add_endline buffer (utf8_of_string (Printf.sprintf " %d" index))
                end)
            local_variable_table;
          List.iter
            (fun (start, length, name, sign, index) ->
              let finish = start + length in
              if (!ofs >= start && !ofs < finish)
                  && not (!prev >= start && !prev < finish) then
                begin
                  utf8_buf_add_string buffer (utf8_of_string (Printf.sprintf "  @LocalVariableTypeTable code%08d: code%08d: " start finish));
                  utf8_buf_add_string buffer name;
                  utf8_buf_add_string buffer space;
                  utf8_buf_add_string buffer (utf8_escape (Signature.utf8_of_field_type_signature sign));
                  utf8_buf_add_endline buffer (utf8_of_string (Printf.sprintf " %d" index))
                end)
            local_variable_type_table;
          let sz, is_wide, mnemo, p, t = Instruction.decompile !ofs i in
          utf8_buf_add_string buffer (utf8_of_string (Printf.sprintf "  code%08d: " !ofs));
          (if is_wide then utf8_buf_add_string buffer wide);
          utf8_buf_add_string buffer (utf8_of_string mnemo);
          List.iter (fun x ->
            utf8_buf_add_string buffer space;
            match x with
            | Instruction.Int_constant ic ->
                utf8_buf_add_string buffer (utf8_of_string (Int64.to_string ic))
            | Instruction.Offset o ->
                utf8_buf_add_string buffer (utf8_of_string (Printf.sprintf "code%08d:" (!ofs + (Int32.to_int o))))
            | Instruction.Float_constant fc ->
                utf8_buf_add_string buffer (utf8_of_string (string_of_float fc))
            | Instruction.String_constant sc ->
                utf8_buf_add_string buffer (utf8_escape sc)
            | Instruction.Class_name cn ->
                utf8_buf_add_string buffer (Name.external_utf8_for_class cn)
            | Instruction.Array_type at ->
                utf8_buf_add_string buffer at
            | Instruction.Primitive_type pt ->
                utf8_buf_add_string buffer (Source.utf8_of_java_type pt)
            | Instruction.Field (cn, fn, d) ->
                utf8_buf_add_string
                  buffer
                  ((Name.external_utf8_for_class cn)
                     ++ dot
                     ++ (Name.utf8_for_field fn)
                     ++ colon
                     ++ (Source.utf8_of_java_type (d :> Descriptor.java_type)))
            | Instruction.Dynamic_method (mn, d) ->
                utf8_buf_add_string
                  buffer
                  (utf8_of_method_call mn d)
            | Instruction.Method (cn, mn, d) ->
                utf8_buf_add_string
                  buffer
                  ((Name.external_utf8_for_class cn)
                     ++ dot
                     ++ (utf8_of_method_call mn d))
            | Instruction.Array_method (at, mn, d) ->
                utf8_buf_add_string
                  buffer
                  ((Source.utf8_of_java_type (at :> Descriptor.java_type))
                     ++ dot
                     ++ (utf8_of_method_call mn d)))
            p;
          (match t with
          | Instruction.No_tail -> ()
          | Instruction.Match_offset_pairs l ->
              List.iter
                (fun (m, o) ->
                  utf8_buf_add_endline buffer empty_utf8;
                  utf8_buf_add_string
                    buffer
                    (utf8_of_string
                       (Printf.sprintf "    %ld => code%08d:" m (!ofs + (Int32.to_int o)))))
                l
          | Instruction.Long_offsets l ->
              List.iter
                (fun o ->
                  utf8_buf_add_endline buffer empty_utf8;
                  utf8_buf_add_string buffer (utf8_of_string (Printf.sprintf "    => code%08d:" (!ofs + (Int32.to_int o)))))
                l);
          utf8_buf_add_endline buffer empty_utf8;
          prev := !ofs;
          ofs := !ofs + sz)
        cv.Attribute.code;
      List.iter
        (fun (start_pc, end_pc, handler_pc, catch_type) ->
          utf8_buf_add_string
            buffer
            (utf8_of_string (Printf.sprintf "  .catch code%08d: code%08d: code%08d:" start_pc end_pc handler_pc));
          utf8_buf_add_endline
            buffer
            (match catch_type with
            | Some n -> space ++ (Name.external_utf8_for_class n)
            | None -> empty_utf8))
        cv.Attribute.exception_table;
      let utf8_of_type = function
        | Attribute.Top_variable_info -> utf8_of_string "top"
        | Attribute.Integer_variable_info -> utf8_of_string "int"
        | Attribute.Float_variable_info -> utf8_of_string "float"
        | Attribute.Long_variable_info -> utf8_of_string "long"
        | Attribute.Double_variable_info -> utf8_of_string "double"
        | Attribute.Null_variable_info -> utf8_of_string "null"
        | Attribute.Uninitialized_this_variable_info -> utf8_of_string "uninit_this"
        | Attribute.Object_variable_info (`Class_or_interface n) -> Name.external_utf8_for_class n
        | Attribute.Object_variable_info (`Array_type at) -> Source.utf8_of_java_type (at :> Descriptor.java_type)
        | Attribute.Uninitialized_variable_info ofs -> utf8_of_string (Printf.sprintf "uninit code%08d:" ofs) in
      List.iter
        (function
          | Attribute.Same_frame ofs ->
              utf8_buf_add_endline
                buffer
                (utf8_of_string (Printf.sprintf "  .frame code%08d: same" ofs))
          | Attribute.Same_locals_1_stack_item_frame (ofs, t) ->
              utf8_buf_add_endline
                buffer
                ((utf8_of_string (Printf.sprintf "  .frame code%08d: same_locals " ofs))
                    ++ (utf8_of_type t))
          | Attribute.Chop_1_frame ofs ->
              utf8_buf_add_endline
                buffer
                (utf8_of_string (Printf.sprintf "  .frame code%08d: chop 1" ofs))
          | Attribute.Chop_2_frame ofs ->
              utf8_buf_add_endline
                buffer
                (utf8_of_string (Printf.sprintf "  .frame code%08d: chop 2" ofs))
          | Attribute.Chop_3_frame ofs ->
              utf8_buf_add_endline
                buffer
                (utf8_of_string (Printf.sprintf "  .frame code%08d: chop 3" ofs))
          | Attribute.Append_1_frame (ofs, t1) ->
              utf8_buf_add_endline
                buffer
                ((utf8_of_string (Printf.sprintf "  .frame code%08d: append " ofs))
                   ++ (utf8_of_type t1))
          | Attribute.Append_2_frame (ofs, t1, t2) ->
              utf8_buf_add_endline
                buffer
                ((utf8_of_string (Printf.sprintf "  .frame code%08d: append " ofs))
                   ++ (utf8_of_type t1) ++ space ++ (utf8_of_type t2))
          | Attribute.Append_3_frame (ofs, t1, t2, t3) ->
              utf8_buf_add_endline
                buffer
                ((utf8_of_string (Printf.sprintf "  .frame code%08d: append " ofs))
                   ++ (utf8_of_type t1) ++ space ++ (utf8_of_type t2) ++ space ++ (utf8_of_type t3))
          | Attribute.Full_frame (ofs, l1, l2) ->
              let utf8_of_type_list l = utf8_concat_sep space (List.map utf8_of_type l) in
              utf8_buf_add_endline
                buffer
                ((utf8_of_string (Printf.sprintf "  .frame code%08d: full " ofs))
                   ++ (utf8_of_type_list l1) ++ tilde ++ (utf8_of_type_list l2)))
        stack_map_table;
      List.iter
        (fun (n, s) ->
          utf8_buf_add_string buffer at_unknown;
          utf8_buf_add_string buffer (utf8_escape n);
          utf8_buf_add_string buffer space;
          utf8_buf_add_endline buffer (utf8_escape (utf8_of_string s)))
        unknowns
  | `Exceptions e ->
      utf8_buf_add_endline
        buffer
        (at_exceptions
           ++ (utf8_concat_sep
                 space
                 (List.map Name.external_utf8_for_class e)))
  | `InnerClasses l ->
      let l' =
        List.map
          (function (inner, outer, name, flags) ->
            at_inner_class
              ++ (match inner with
              | Some cn -> Name.external_utf8_for_class cn
              | None -> zero)
              ++ space
              ++ (match outer with
              | Some cn -> Name.external_utf8_for_class cn
              | None -> zero)
              ++ space
              ++ (match name with
              | Some n -> n
              | None -> zero)
              ++ space
              ++ (Source.utf8_of_flags (flags :> AccessFlag.t list)))
          l in
      utf8_buf_add_endline
        buffer
        (utf8_concat_sep (utf8_of_string "\n  ") l')
  | `EnclosingMethod (cn, m) ->
      utf8_buf_add_string buffer at_enclosing_method;
      utf8_buf_add_string buffer (Name.external_utf8_for_class cn);
      (match m with
      | Some (n, d) ->
          utf8_buf_add_string buffer space;
          utf8_buf_add_endline buffer (utf8_of_method_call n d)
      | None -> utf8_buf_add_endline buffer empty_utf8)
  | `Synthetic ->
      utf8_buf_add_endline buffer at_synthetic
  | `Signature s ->
      let s' = (match s with
      | `Class cs -> Signature.utf8_of_class_signature cs
      | `Method ms -> Signature.utf8_of_method_signature ms
      | `Field fs -> Signature.utf8_of_field_type_signature fs) in
      utf8_buf_add_endline
        buffer
        (at_signature ++ (utf8_escape s'))
  | `SourceFile sf ->
      utf8_buf_add_endline
        buffer
        (at_source_file ++ (utf8_escape sf))
  | `SourceDebugExtension sde ->
      utf8_buf_add_endline
        buffer
        (at_source_debug_extension ++ (utf8_escape sde))
  | `LineNumberTable _ -> ()
  | `LocalVariableTable _ -> ()
  | `LocalVariableTypeTable _ -> ()
  | `Deprecated ->
      utf8_buf_add_endline buffer at_deprecated
  | `RuntimeVisibleAnnotations l ->
      let l' = List.flatten
          (List.map
             (annotation_list at_runtime_visible_annotations)
             l) in
      utf8_buf_add_endline buffer (utf8_concat_sep (utf8_of_string "\n  ") l')
  | `RuntimeInvisibleAnnotations l ->
      let l' = List.flatten
          (List.map
             (annotation_list at_runtime_invisible_annotations)
             l) in
      utf8_buf_add_endline buffer (utf8_concat_sep (utf8_of_string "\n  ") l')
  | `RuntimeVisibleParameterAnnotations l ->
      let no = ref 0 in
      List.iter
        (fun x ->
          let l' = List.flatten
              (List.map
                 (annotation_list (utf8_of_string ((if !no = 0 then "" else "  ") ^ "@RuntimeVisibleParameterAnnotations " ^ (string_of_int !no))))
                 x) in
          utf8_buf_add_endline buffer (utf8_concat_sep (utf8_of_string "\n  ") l');
          incr no)
        l
  | `RuntimeInvisibleParameterAnnotations l ->
      let no = ref 0 in
      List.iter
        (fun x ->
          let l' = List.flatten
              (List.map
                 (annotation_list (utf8_of_string ((if !no = 0 then "" else "  ") ^ "@RuntimeInvisibleParameterAnnotations " ^ (string_of_int !no))))
                 x) in
          utf8_buf_add_endline buffer (utf8_concat_sep (utf8_of_string "\n  ") l');
          incr no)
        l
  | `AnnotationDefault ad ->
      let l = annotation_value_list at_annotation_default ad in
      utf8_buf_add_endline buffer (utf8_concat_sep (utf8_of_string "\n  ") l)
  | `StackMapTable _ -> ()
  | `Unknown (n, s) ->
      utf8_buf_add_string buffer at_unknown;
      utf8_buf_add_string buffer (utf8_escape n);
      utf8_buf_add_string buffer space;
      utf8_buf_add_endline buffer (utf8_escape (utf8_of_string s))

let disassemble_to_buffer buffer cp s =
  let cl = ClassLoader.make cp in
  let cd = ClassLoader.find cl s in
  utf8_buf_add_endline
    buffer
    (dot_class
       ++ (Source.utf8_of_flags (cd.ClassDefinition.access_flags :> AccessFlag.t list))
       ++ (Name.external_utf8_for_class cd.ClassDefinition.name));
  (match cd.ClassDefinition.extends with
  | Some c ->
      utf8_buf_add_endline
        buffer
        (dot_extends ++ (Name.external_utf8_for_class c))
  | None -> ());
  let interfaces = List.map Name.external_utf8_for_class cd.ClassDefinition.implements in
  List.iter
    (fun x ->
      utf8_buf_add_endline
        buffer
        (dot_implements ++ x))
    (List.sort utf8_compare interfaces);
  List.iter
    (add_attribute buffer)
    (List.sort
       Source.attribute_order
       (cd.ClassDefinition.attributes :> Attribute.t list));
  if cd.ClassDefinition.fields <> [] then
    utf8_buf_add_endline buffer empty_utf8;
  List.iter
    (fun (flags, name, desc, attrs) ->
        utf8_buf_add_endline buffer empty_utf8;
        utf8_buf_add_endline
          buffer
          (dot_field
             ++ (Source.utf8_of_flags (List.sort Source.flag_order (flags :> AccessFlag.t list)))
             ++ (Source.utf8_of_java_type (desc :> Descriptor.java_type))
             ++ space
             ++ (Name.utf8_for_field name));
        List.iter
          (add_attribute buffer)
          (List.sort
             Source.attribute_order
             (attrs :> Attribute.t list)))
    (List.sort Source.field_order cd.ClassDefinition.fields);
  if cd.ClassDefinition.methods <> [] then
    utf8_buf_add_endline buffer empty_utf8;
  List.iter
    (fun x ->
      let flags, name, desc, attrs =
        match x with
        | Method.Regular (f, n, d, a) -> (f, n, d, a)
        | Method.Constructor (f, d, a) -> ((f :> AccessFlag.for_method list), class_constructor_name, (d, `Void), a)
        | Method.Initializer (s, a) -> ((if s then [`Strict] else []), class_initializer_name, ([], `Void), a) in
      utf8_buf_add_endline buffer empty_utf8;
      utf8_buf_add_endline
        buffer
        (dot_method
           ++ (Source.utf8_of_flags (flags :> AccessFlag.t list))
           ++ (utf8_of_method_desc name desc));
      List.iter
        (add_attribute buffer)
        (List.sort
           Source.attribute_order
           (attrs :> Attribute.t list)))
    (List.sort Source.method_order cd.ClassDefinition.methods)

let disassemble_to_channel chan cp s =
  let buffer = utf8_buf_make () in
  disassemble_to_buffer buffer cp s;
  Source.output chan buffer

let disassemble cp s =
  disassemble_to_channel stdout cp s
