/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    private static final MethodRef println =
	new MethodRef("java.io.PrintStream",
		      "println",
		      Descriptor.VOID,
		      descs(new Descriptor.Class("java.lang.String")));

    public static void main(final String[] args) throws Throwable {
	final List<Instruction> catchInstructions =
	    asList(new NEW(new TypeName("java.lang.StringBuilder")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.StringBuilder",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new CHECKCAST_ARRAY(new Descriptor.Array(Descriptor.INT)),
		   new LDC_STRING("uncaught"),
		   new GOTO((short) 5),
		   new LDC_STRING("caught"),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> catchNpeInstructions =
	    asList(ACONST_NULL,
		   new INVOKEVIRTUAL(new MethodRef("java.lang.Object",
						   "toString",
						   new Descriptor.Class("java.lang.String"),
						   descs())),
		   new LDC_STRING("uncaught"),
		   new GOTO((short) 13),
		   new LDC_STRING("caught"),
		   new GOTO((short) 8),
		   new LDC_STRING("wrong caught"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> throwInstructions =
	    asList(new NEW(new TypeName("java.lang.ClassCastException")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.ClassCastException",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   ATHROW,
		   new LDC_STRING("uncaught"),
		   new GOTO((short) 5),
		   new LDC_STRING("caught"),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> finallyInstructions =
	    asList(ACONST_NULL,
		   new INVOKEVIRTUAL(new MethodRef("java.lang.Object",
						   "toString",
						   new Descriptor.Class("java.lang.String"),
						   descs())),
		   new JSR((short) 8),
		   RETURN,
		   new JSR((short) 4),
		   RETURN,
		   ASTORE_1,
		   new GETSTATIC(systOut),
		   new LDC_STRING("finally"),
		   new INVOKEVIRTUAL(println),
		   new RET((short) 1),
		   RETURN);
	final List<Instruction> finallyWInstructions =
	    asList(ACONST_NULL,
		   new INVOKEVIRTUAL(new MethodRef("java.lang.Object",
						   "toString",
						   new Descriptor.Class("java.lang.String"),
						   descs())),
		   new JSR_W(12),
		   RETURN,
		   new JSR_W(6),
		   RETURN,
		   ASTORE_1,
		   new GETSTATIC(systOut),
		   new LDC_STRING("finally"),
		   new INVOKEVIRTUAL(println),
		   new RET((short) 1),
		   RETURN);
	final List<Instruction> instructions =
	    asList(new INVOKESTATIC(new MethodRef("pack.Test",
						  "catch",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "catch_npe",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "throw",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "finally",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "finally_w",
						  Descriptor.VOID,
						  descs())),
		   RETURN);
	final Method catch_ =
	    compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
			  "catch",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  asList(new Attribute.ExceptionTableElement(0, 15, 15, null)),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  catchInstructions);
	final Method catch_npe =
	    compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
			  "catch_npe",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  asList(new Attribute.ExceptionTableElement(0, 9, 9, "java.lang.NullPointerException"),
				 new Attribute.ExceptionTableElement(0, 9, 14, "java.lang.ArrayIndexOutOfBoundsException")),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  catchNpeInstructions);
	final Method throw_ =
	    compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
			  "throw",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  asList(new Attribute.ExceptionTableElement(0, 13, 13, null)),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  throwInstructions);
	final Method finally_ =
	    compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
			  "finally",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  asList(new Attribute.ExceptionTableElement(0, 8, 8, null)),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  finallyInstructions);
	final Method finally_w =
	    compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
			  "finally_w",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  asList(new Attribute.ExceptionTableElement(0, 10, 10, null)),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  finallyWInstructions);
	final Method main = compileMethod(instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						null,
						null,
						asList(catch_,
						       catch_npe,
						       throw_,
						       finally_,
						       finally_w,
						       main));
	final byte[] bc = ByteCode.encode(Version.Java_1_5, cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
