/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    private static final MethodRef println =
	new MethodRef("java.io.PrintStream",
		      "println",
		      Descriptor.VOID,
		      descs(new Descriptor.Class("java.lang.String")));

    public static void main(final String[] args) throws Throwable {
	final List<Instruction> resultInstructions =
	    asList(ILOAD_0,
		   new LOOKUPSWITCH(50, 3, asList(new MatchOffsetPair(-1, 35),
						  new MatchOffsetPair(0, 40),
						  new MatchOffsetPair(1, 45))),
		   new LDC_STRING("lower"),
		   new GOTO((short) 18),
		   new LDC_STRING("equal"),
		   new GOTO((short) 13),
		   new LDC_STRING("greater"),
		   new GOTO((short) 8),
		   new LDC_STRING("default"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> doubleInstructions =
	    asList(new LDC2_W_DOUBLE(1.2),
		   new LDC2_W_DOUBLE(3.4),
		   DCMPG,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   new LDC2_W_DOUBLE(5.6),
		   new LDC2_W_DOUBLE(3.4),
		   DCMPL,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   RETURN);
	final List<Instruction> floatInstructions =
	    asList(new LDC_FLOAT(1.2f),
		   new LDC_FLOAT(1.2f),
		   FCMPG,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   new LDC_FLOAT(5.6f),
		   new LDC_FLOAT(3.4f),
		   FCMPL,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   RETURN);
	final List<Instruction> longInstructions =
	    asList(new LDC2_W_LONG(12L),
		   new LDC2_W_LONG(34L),
		   LCMP,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   new LDC2_W_LONG(56L),
		   new LDC2_W_LONG(34L),
		   LCMP,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "result",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   RETURN);
	final List<Instruction> instanceofInstructions =
	    asList(new NEW(new TypeName("java.lang.String")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.String",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new INSTANCEOF_TYPE(new TypeName("java.lang.Integer")),
		   new IFNE((short) 8),
		   new LDC_STRING("not inst"),
		   new GOTO((short) 5),
		   new LDC_STRING("inst"),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> nullInstructions =
	    asList(ALOAD_0,
		   new IFNULL((short) 10),
		   ALOAD_0,
		   new IFNONNULL((short) 11),
		   new GOTO((short) 13),
		   new LDC_STRING("null"),
		   new GOTO((short) 13),
		   new LDC_STRING("non null"),
		   new GOTO((short) 8),
		   new LDC_STRING("error"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> refEqInstructions =
	    asList(ALOAD_0,
		   ALOAD_1,
		   new IF_ACMPEQ((short) 13),
		   ALOAD_0,
		   ALOAD_1,
		   new IF_ACMPNE((short) 13),
		   new LDC_STRING("error"),
		   new GOTO((short) 13),
		   new LDC_STRING("equal references"),
		   new GOTO((short) 8),
		   new LDC_STRING("different references"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> intEqInstructions =
	    asList(ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPEQ((short) 33),
		   ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPNE((short) 33),
		   ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPLT((short) 33),
		   ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPLE((short) 33),
		   ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPGT((short) 33),
		   ILOAD_0,
		   ILOAD_1,
		   new IF_ICMPGE((short) 33),
		   new LDC_STRING("error"),
		   new GOTO((short) 33),
		   new LDC_STRING("eq"),
		   new GOTO((short) 28),
		   new LDC_STRING("ne"),
		   new GOTO((short) 23),
		   new LDC_STRING("lt/never reached"),
		   new GOTO((short) 18),
		   new LDC_STRING("le/never reached"),
		   new GOTO((short) 13),
		   new LDC_STRING("gt/never reached"),
		   new GOTO((short) 8),
		   new LDC_STRING("ge/never reached"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> intEq0Instructions =
	    asList(ILOAD_0,
		   new IFEQ((short) 28),
		   ILOAD_0,
		   new IFNE((short) 29),
		   ILOAD_0,
		   new IFLT((short) 30),
		   ILOAD_0,
		   new IFLE((short) 31),
		   ILOAD_0,
		   new IFGT((short) 32),
		   ILOAD_0,
		   new IFGE((short) 33),
		   new LDC_STRING("error"),
		   new GOTO((short) 33),
		   new LDC_STRING("eq"),
		   new GOTO((short) 28),
		   new LDC_STRING("ne"),
		   new GOTO((short) 23),
		   new LDC_STRING("lt/never reached"),
		   new GOTO((short) 18),
		   new LDC_STRING("le/never reached"),
		   new GOTO((short) 13),
		   new LDC_STRING("gt/never reached"),
		   new GOTO((short) 8),
		   new LDC_STRING("ge/never reached"),
		   new GOTO((short) 3),
		   new GETSTATIC(systOut),
		   SWAP,
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final List<Instruction> instructions =
	    asList(new INVOKESTATIC(new MethodRef("pack.Test",
						  "double",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "float",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "long",
						  Descriptor.VOID,
						  descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "instanceof",
						  Descriptor.VOID,
						  descs())),

		   ACONST_NULL,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "null",
						  Descriptor.VOID,
						  descs(new Descriptor.Class("java.lang.Object")))),

		   new NEW(new TypeName("java.lang.String")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.String",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "null",
						  Descriptor.VOID,
						  descs(new Descriptor.Class("java.lang.Object")))),

		   ACONST_NULL,
		   new NEW(new TypeName("java.lang.String")),
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.String",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "ref_eq",
						  Descriptor.VOID,
						  descs(new Descriptor.Class("java.lang.Object"),
							new Descriptor.Class("java.lang.Object")))),

		   new NEW(new TypeName("java.lang.String")),
		   DUP,
		   DUP,
		   new INVOKESPECIAL(new MethodRef("java.lang.String",
						   "<init>",
						   Descriptor.VOID,
						   descs())),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "ref_eq",
						  Descriptor.VOID,
						  descs(new Descriptor.Class("java.lang.Object"),
							new Descriptor.Class("java.lang.Object")))),

		   new LDC_INT(3),
		   new LDC_INT(4),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq",
						  Descriptor.VOID,
						  descs(Descriptor.INT,
							Descriptor.INT))),

		   new LDC_INT(3),
		   new LDC_INT(-4),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq",
						  Descriptor.VOID,
						  descs(Descriptor.INT,
							Descriptor.INT))),

		   new LDC_INT(3),
		   new LDC_INT(3),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq",
						  Descriptor.VOID,
						  descs(Descriptor.INT,
							Descriptor.INT))),

		   new LDC_INT(4),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq0",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),

		   new LDC_INT(-4),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq0",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),

		   new LDC_INT(0),
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "int_eq0",
						  Descriptor.VOID,
						  descs(Descriptor.INT))),
		   RETURN);
	final Method result =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "result",
			  Descriptor.VOID,
			  descs(Descriptor.INT),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  resultInstructions);
	final Method double_ =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "double",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  doubleInstructions);
	final Method float_ =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "float",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  floatInstructions);
	final Method long_ =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "long",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  longInstructions);
	final Method instanceof_ =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "instanceof",
			  Descriptor.VOID,
			  descs(),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  instanceofInstructions);
	final Method null_ =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "null",
			  Descriptor.VOID,
			  descs(new Descriptor.Class("java.lang.Object")),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  nullInstructions);
	final Method refEq =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "ref_eq",
			  Descriptor.VOID,
			  descs(new Descriptor.Class("java.lang.Object"),
				new Descriptor.Class("java.lang.Object")),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  refEqInstructions);
	final Method intEq =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "int_eq",
			  Descriptor.VOID,
			  descs(Descriptor.INT, Descriptor.INT),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  intEqInstructions);
	final Method intEq0 =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "int_eq0",
			  Descriptor.VOID,
			  descs(Descriptor.INT),
			  4,
			  4,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  intEq0Instructions);
	final Method main = compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
					  "main",
					  Descriptor.VOID,
					  descs(new Descriptor.Array(new Descriptor.Class("java.lang.String"))),
					  5,
					  4,
					  Collections.<Attribute.ExceptionTableElement>emptyList(),
					  Collections.<Attribute>emptyList(),
					  Collections.<Attribute>emptyList(),
					  instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						null,
						null,
						asList(result,
						       double_,
						       float_,
						       long_,
						       instanceof_,
						       null_,
						       refEq,
						       intEq,
						       intEq0,
						       main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
